from yo_ds__tests.common import *

class TestOtherPlots(TestCase):
    def test_pie(self):
        series = Query.en(range(1,5)).to_series()
        pie_plot(series)

    def test_roc_auc(self):
        true = [0,0,0,0,0,1,1,1]
        pred = [0.1,0.3,0.5,0.7,0.4,0.6,0.8, 0.8]
        return roc_plot(true,pred)


from yo_ds__tests.test_plots.test_fluent_plot import TestFluentPlot
from yo_ds__tests.test_plots.test_grbar import TestGrBar
from yo_ds__tests.test_plots.test_other_plots import TestOtherPlots
from yo_ds__tests.common import *



## TODO: this file is currently invalid, because no tests are available

class PlotDemoTest(TestCase):

    def make_nbook_section(self, filename, testtype, header):
        this_file_name = path('test_plots',filename)
        lines = Query.file.text(this_file_name).to_list()
        insertion = Query.en(lines).with_indices().where(lambda z: not z.value.startswith('from')).first().key
        lines.insert(insertion, 'class TestCase:')
        lines.insert(insertion + 1, '    pass')
        lines = Query.en(lines).select(lambda z: z + "\n").to_list()
        lines.append('tc = {0}()'.format(testtype.__name__))

        cells = [MD_CELL_TEMPLATE(['# '+header]), CELL_TEMPLATE(lines)]

        ts = testtype()
        Query.en(dir(ts)).where(lambda z: z.startswith('test')).select('tc.{0}()'.format).select(CELL_TEMPLATE).foreach(cells.append)

        return cells



    def test_mkbook(self):

        c1 = self.make_nbook_section('test_fluent_plot.py', TestFluentPlot, 'Fluent plot examples')
        c2 = self.make_nbook_section('test_grbar.py', TestGrBar, 'GRBAR plot examples')
        c3 = self.make_nbook_section('test_other_plots.py',TestOtherPlots, 'Other plots examples')


        nb = NOTEBOOK_TEMPLATE(c1+c2+c3)
        FileIO.write_json(nb,FileIO.relative_to_file(__file__,'Plots demo.ipynb'))


def NOTEBOOK_TEMPLATE(cells):
    return {
        "cells": cells,
        "metadata": {
            "kernelspec": {
                "display_name": "Python 3",
                "language": "python",
                "name": "python3"
            },
            "language_info": {
                "codemirror_mode": {
                    "name": "ipython",
                    "version": 3
                },
                "file_extension": ".py",
                "mimetype": "text/x-python",
                "name": "python",
                "nbconvert_exporter": "python",
                "pygments_lexer": "ipython3",
                "version": "3.6.7"
            }
        },
        "nbformat": 4,
        "nbformat_minor": 2
    }


def CELL_TEMPLATE(lines):
    return {
        "cell_type": "code",
        "execution_count": 1,
        "metadata": {},
        "outputs": [],
        "source": lines
    }

def MD_CELL_TEMPLATE(lines):
    return {
   "cell_type": "markdown",
   "metadata": {},
   "source": lines
  }



import unittest
from yo_ds__tests.common import *
import pandas as pd
from yo_ds.metrics import Metrics, roc_optimal_threshold

uniclass = pd.DataFrame(dict(
    true = [0,0,0,0,0,1,1,1,1,1],
    predicted = [0,0,0,1,1,1,1,0,0,0]
))

multiclass = pd.DataFrame(dict(
    true = [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2],
    predicted = [0, 0, 1, 2,  1, 2, 0, 0, 0, 1, 0, 2, 2, 2]
))

roc = pd.DataFrame(dict(
    true = [0, 0, 0, 0, 1, 1, 1],
    predicted = [0.1, 0.2, 0.3, 0.6, 0.5, 0.8, 0.9]
))

class MetricsTests(unittest.TestCase):
    def test_uniclass(self):
        series = Metrics.Uniclass.compute(uniclass.true, uniclass.predicted).to_series()

    def test_multiclass_matrix(self):
        matrix = Metrics.Multiclass.by_class_matrix(multiclass.true,multiclass.predicted).to_dataframe().sort_values('metric_name')

    def test_multiclass(self):
        metrics = Metrics.Multiclass.compute(multiclass.true, multiclass.predicted).to_dataframe().drop('metric_type',axis=1)

    def test_roc_auc(self):
        threshold = roc_optimal_threshold(roc.true,roc.predicted)
        self.assertEqual(0.5,threshold)


from yo_extensions import *
from yo_extensions.misc import *
from yo_extensions import fluq as fluq
from unittest import TestCase

class CompatibilityTestCase(TestCase):
    def test_query(self):
        df = Query.combinatorics.grid(a=[10],b=[20]).to_dataframe()
        pass

    def test_query_feed(self):
        path = IO.relative_to_file(__file__,'sample.txt')
        Query.args(1,2,3).select(str).feed(fluq.to_text_file(path))

    def test_query_inherited_feed(self):
        self.assertListEqual(
            [(1,2),(2,3)],
            Query.args(1,2,3).feed(fluq.pairwise()).to_list()
        )


    def test_df_alg(self):
        s = Query.en(range(10)).to_series().feed(alg.trimmer(drop_before=5))
        self.assertListEqual(
            [6,7,8,9],
            list(s.values)
        )







import unittest
from yo_ds import kraken
from collections import OrderedDict
from sklearn.linear_model import Lasso, Ridge
from yo_ds__tests.common import *
import shutil
import os
from typing import *

def kraken_default_method(iteration, a, b):
    return pd.DataFrame(dict(result=pd.Series([a+b,a-b]), is_addition=[True,False]))

def kraken_simple_method(iteration, a, b):
    return a+b

def kraken_throwing_method(iteration, a,  b):
    if iteration==3:
        raise ValueError("I don't like this number")
    return a+b

def kraken_nested_method(iteration, a):
    return pd.DataFrame(dict(result=pd.Series([a.b+a.c])))

class KrakenTests(unittest.TestCase):

    #this method is required to override default value to with_tqdm and pandas_extractor
    def kraken_run(self, method, plan, **kwargs):
        kwargs['with_tqdm'] = maybe(kwargs, ['with_tqdm'], default=False)
        kwargs['pandas_extractor'] = maybe(kwargs,['pandas_extractor'])
        return kraken.release(method, plan, **kwargs)


    def assertResult(self, result: List[Any], kraken_output: List[kraken.IterationResult]):
        actual = [z.result for z in kraken_output]
        self.assertListEqual(result,actual)
        for r in kraken_output:
            self.assertIsInstance(r.condition,OrderedDict)
            self.assertTrue(r.status == kraken.IterationStatus.Success)

    def get_default_plan(self):
        return Query.combinatorics.grid(a=[10,20], b=[1,2]).to_list()


    #the default call to kraken. Each iteration returns Dataframe, and they are later merged with conditions
    def test_default_machinery(self):
        plan = self.get_default_plan()
        df = self.kraken_run(
            kraken_default_method,
            plan,
            pandas_extractor = lambda df, _ : df #the default one. Normally you don't need to override it
        )

        self.assertListEqual([11,9,12,8,21,19,22,18],list(df.result))
        self.assertListEqual([True,False,True,False,True,False,True,False], list(df.is_addition))
        self.assertListEqual([0,0,1,1,2,2,3,3], list(df.iteration))
        self.assertListEqual([10,10,10,10,20,20,20,20],list(df.a))
        self.assertListEqual([1,1,2,2,1,1,2,2],list(df.b))

    # kraken method can return anything.
    # If this case, you can skip the conversion to pandas and process raw outputs/configs
    def test_without_extractor(self):
        plan = self.get_default_plan()
        result = self.kraken_run(
            kraken_simple_method,
            plan,
            pandas_extractor=None)

        self.assertResult([11,12,21,22],result)


    # you can also provide the custom extractor, that will transform method's output to dataframe
    # this method also accepts the condition, but DO NOT MERGE manually! It's done later automatically. It's mostly for the sake of generality
    def test_with_custom_extractor(self):
        plan = self.get_default_plan()
        result = self.kraken_run(
            kraken_simple_method,
            plan,
            pandas_extractor=lambda z, _:pd.DataFrame(dict(result=pd.Series([z]))))
        self.assertIsInstance(result,pd.DataFrame)
        self.assertListEqual([11,12,21,22],list(result.result))
        self.assertListEqual([0,1,2,3],list(result.iteration))
        self.assertListEqual([10,10,20,20], list(result.a))

    # you can run the process on the specific iterations. It is helpful if some of them raised exception
    def test_with_specific_iteration(self):
        plan = self.get_default_plan()
        result = self.kraken_run(
            kraken_simple_method,
            plan,
            special_iterations=[1,3]
        )
        self.assertResult([12, 22], result)


    # By default, if the method throws, the kraken do not handles it
    def test_with_throwing_method_0(self):
        plan = self.get_default_plan()
        self.assertRaises(ValueError,lambda: self.kraken_run(kraken_throwing_method,plan))

    # But if handle_exception_callback is provided, each erroneous exception will be returned for postprocessing,
    # and the erroneous entries will be stored in the resulting report
    def test_with_throwing_method_1(self):
        plan = self.get_default_plan()
        errors = []
        result = self.kraken_run(
            kraken_throwing_method,
            plan,
            handle_exception_callback = lambda z: errors.append(z)
        )
        self.assertEqual(kraken.IterationStatus.Failed, result[3].status)
        self.assertEqual(3,errors[0].iteration)
        good_result = result[0:3]
        self.assertResult([11,12,21],good_result)

    #Erroneous entries are excluded in dataframe report
    def test_with_throwing_method_2(self):
        plan = self.get_default_plan()
        errors = []
        result = self.kraken_run(
            kraken_throwing_method,
            plan,
            handle_exception_callback=lambda z: errors.append(z),
            pandas_extractor = lambda res, _ : pd.DataFrame(pd.Series([res]))
        )
        self.assertListEqual([0,1,2],list(result.iteration))

    # Plan can contain nested fields.
    # It's useful if you want to, e.g., independently test featurizers and models
    # The method will take arguments exactly as specified in the plan. But in the resulting dataframe, the columns will be flattened
    def test_with_nested_field(self):
        plan = Query.combinatorics.grid(a = Query.combinatorics.grid(b=[1,2],c=[10,20]).to_list()).to_list()
        result = self.kraken_run(
            kraken_nested_method,
            plan,
            pandas_extractor=lambda df,_:df
        )
        self.assertListEqual([11,21,12,22], list(result.result))
        self.assertListEqual(['result','iteration','a_b','a_c'],list(result.columns))

    # If you want to test different models, you can make use of kraken.invoke method.
    # To do it, set 'ctor' field to constructor, additional fields in the corresponding constructors arguments, select via kraken.invoke and obtain object from 'instance'
    def test_invoke(self):
        models = Query.combinatorics.grid(ctor=[Lasso,Ridge], alpha=[1,2]).select(kraken.invoke).to_list()
        self.assertIsInstance(models[0]['instance'],Lasso)
        self.assertEqual(models[0]['instance'].alpha,1)
        self.assertIsInstance(models[3]['instance'],Ridge)
        self.assertEqual(models[3]['instance'].alpha, 2)

    # Kraken can run in parallel. The test does not check if it was actually run in parallel, just that the method is working
    def test_parallel(self):
        plan = self.get_default_plan()
        result = self.kraken_run(
            kraken_simple_method,
            plan,
            parallel_kwargs = {}
        )


    # Instead of immediately processing result, Kraken can store them in files. Kraken do noe re-evaluate stored results
    # The stored data can be retrieved in the same format as Kraken return with the method load
    def test_files(self):
        folder = path('kraken_output_files/')
        print(folder)
        if os.path.isdir(folder): #pragma: no cover
            shutil.rmtree(folder)

        plan = self.get_default_plan()

        self.kraken_run(
            kraken_simple_method,
            plan,
            cache_to_folder=folder,
            special_iterations=[3]
        )

        file = Query.folder(folder).single()
        self.assertEqual('3.kraken.pkl',file.name)

        results = self.kraken_run(
            kraken_simple_method,
            plan,
            cache_to_folder=folder
        ) # type: List[kraken.IterationResult]

        for index,result in enumerate(results):
            self.assertEqual(
                kraken.IterationStatus.Skipped if index == 3 else kraken.IterationStatus.Success,
                result.status)
            self.assertIsNone(result.condition)
            self.assertIsNone(result.result)

        loaded_results = Query.en(kraken.load(folder, None)).order_by(lambda z: z.result).to_list()
        self.assertResult([11,12,21,22],loaded_results)

        shutil.rmtree(folder)

    #You can also shuffle result. It's a good option if different entries of the plan require different time,
    #but you want to have more accurate time estimation from tqdm
    def test_shuffle(self):
        plan = self.get_default_plan()
        result = self.kraken_run(
            kraken_simple_method,
            plan,
            shuffle=1
        )
        self.assertResult([22,21,11,12],result)


from  unittest import TestCase
import os
from yo_ds import *


def path(*args):
    folder = FileIO.find_root_folder('yo.root')
    temp = os.path.join(folder, 'yo_ds__tests', *args)
    os.makedirs(os.path.dirname(temp), exist_ok=True)
    return temp




from typing import *
from yo_fluq_ds import *
from sklearn.metrics import precision_score, recall_score, accuracy_score, f1_score, roc_auc_score
import numpy as np

class MulticlassMetricType(OrderedEnum):
    General = 0
    ByClass = 1
    Global = 2

class MulticlassMetrics(Obj):
    def __init__(self,
                 metric_type: MulticlassMetricType,
                 metric_name: str,
                 metric_aggregation: Optional[str],
                 value: float):
        super(MulticlassMetrics, self).__init__()
        self.metric_type = metric_type
        self.metric_name = metric_name
        self.metric_aggregation = metric_aggregation
        self.value = value

class MulticlassByClassMetrics(Obj):
    def __init__(self, class_name: Any, metric_name: str, value: float):
        super(MulticlassByClassMetrics, self).__init__()
        self.class_name = class_name
        self.metric_name = metric_name
        self.value = value


class MulticlassMetricsComputer:
    def __init__(self,
                 global_metrics : Dict[str,Callable[[pd.Series,pd.Series],float]],
                 general_metrics: Dict[str,Callable[[pd.Series,pd.Series,str],float]],
                 general_aggregators: List[str],
                 by_class_metrics: Dict[str, Callable[[pd.Series, pd.Series], float]],
                 by_class_aggregators: Dict[str, Callable[[pd.Series], float]]
                 ):
        self.global_metrics = global_metrics
        self.by_class_metrics = by_class_metrics
        self.by_class_aggregators = by_class_aggregators
        self.general_metrics = general_metrics
        self.general_aggregators = general_aggregators

    def _get_global(self, true, predicted)->Queryable[MulticlassMetrics]:
        return (Query
                .dict(self.global_metrics)
                .select(lambda metric: MulticlassMetrics(
                    MulticlassMetricType.Global,
                    metric.key,
                    None,
                    metric.value(true,predicted)
                )))

    def _get_general(self, true, predicted) -> Queryable[MulticlassMetrics]:
        return (Query
                .en(self.general_aggregators)
                .select_many(lambda agg: Query
                             .dict(self.general_metrics)
                             .select(lambda metric: MulticlassMetrics(
                                    MulticlassMetricType.General,
                                    metric.key,
                                    agg,
                                    metric.value(true,predicted,agg)
                ))))

    def by_class_matrix(self, true, predicted) -> Queryable[MulticlassByClassMetrics]:
        classes = Query.en(true).intersect(predicted).distinct().to_list()
        return (Query
                .en(classes)
                .select_many(lambda _class: Query
                             .dict(self.by_class_metrics)
                             .select(lambda metric: MulticlassByClassMetrics(
                                    _class,
                                    metric.key,
                                    metric.value(true==_class,predicted==_class)
                ))))

    def _get_by_class(self, true, predicted) -> Queryable[MulticlassMetrics]:
        return (self.by_class_matrix(true,predicted)
                .group_by(lambda z: z.metric_name)
                .select_many(lambda group: Query
                             .dict(self.by_class_aggregators)
                             .select(lambda agg: MulticlassMetrics(
                                    MulticlassMetricType.ByClass,
                                    group.key,
                                    agg.key,
                                    agg.value([v.value for v in group])
                ))))


    def compute(self, true, predicted):
        return (self._get_general(true,predicted)
                .concat(self._get_global(true,predicted))
                .concat(self._get_by_class(true, predicted))
                )

class UniclassMetricsComputer:
    def __init__(self, metrics: Dict[str,Callable[[pd.Series,pd.Series],float]]):
        self.metrics = metrics

    def compute(self, true, predicted)->Queryable[KeyValuePair[str,float]]:
        return (Query
                .dict(self.metrics)
                .select(lambda metric: KeyValuePair(metric.key,metric.value(true,predicted)))
        )

def _wrap_general(m):
    return lambda true, predicted, agg : m(true,predicted, average = agg)

general_metrics= Obj(
    precision = _wrap_general(precision_score),
    recall = _wrap_general(recall_score),
    f1 = _wrap_general(f1_score)
)

general_aggregators = [
    'macro',
    'micro',
    'weighted'
]

by_class_metrics = Obj(
    precision = precision_score,
    recall = recall_score,
    f1 = f1_score,
    accuracy = accuracy_score
)

by_class_aggregators = Obj(
    min = lambda z: min(z),
    max = lambda z: max(z),
    mean = lambda z: np.mean(z),
    std = lambda z: np.std(z)
)

global_metrics = Obj(
    accuracy = accuracy_score
)

class Metrics:
    Uniclass = UniclassMetricsComputer(by_class_metrics)
    Multiclass = MulticlassMetricsComputer(global_metrics,general_metrics,general_aggregators,by_class_metrics,by_class_aggregators)



from ._common import *

def pie_plot(series: pd.Series, explode: float = 0.1, with_abs_values: bool = False, ax: Axes = None):
    def pcnt(p, total):
        cnt = round(p * total / 100)
        return "{0:.0f} ({1:.2f}%)".format(cnt, p)

    values = series.name or 'values'
    ind = series.index.name or 'index'
    dt = series.to_frame(values).reset_index().assign(explode=explode)


    total = dt[values].sum()
    pcnter = '%1.2f%%'
    if with_abs_values:
        pcnter = lambda p: pcnt(p, total)

    ax = default_ax(ax)

    ax.pie(
        dt[values],
        labels=dt[ind],
        explode=dt['explode'],
        shadow=True,
        startangle=90,
        autopct=pcnter
    )


from yo_ds__tests.test_plots.common import *

class TestPlotsMisc(TestCase):
    def test_fluentplot_iterate_df_columns_raises(self):
        df = pd.Series([1,2,3])
        self.assertRaises(
            ValueError,
            lambda: df.feed(FluentPlot().call(lambda z, ax: z.plot(ax=ax)).iterate_df_columns())
        )


from typing import *
import pandas as pd
from matplotlib.axes import Axes
from matplotlib import pyplot as plt

class FluentPlot:
    def __init__(self):
        super(FluentPlot, self).__init__()

        self._default_ax = None  # type: Optional[Axes]
        self._default_fig_size = None  # type: Optional[Tuple[float,float]]
        self._post_processing = []  # type: List[Callable[[Axes],Any]]

        self._title = None  # type: Optional[str]
        self._xlabel = None  # type: Optional[str]
        self._ylabel = None  # type: Optional[str]
        self._with_legend = False  # type: bool

        self._method = None
        self._iterate = False
        self._iterate_df_columns = False


    def __call__(self, obj: Any) -> Axes:
        if self._default_ax is not None:
            ax = self._default_ax
        elif self._default_fig_size is not None:
            _, ax = plt.subplots(1, 1, figsize=self._default_fig_size)
        else:
            _, ax = plt.subplots(1, 1, figsize=(12, 6))

        self._draw(obj, ax)

        if self._title is not None:
            ax.set_title(self._title)
        if self._xlabel is not None:
            ax.set_xlabel(self._xlabel)
        if self._ylabel is not None:
            ax.set_ylabel(self._ylabel)
        if self._with_legend:
            ax.legend()

        for post_proc in self._post_processing:
            post_proc(ax)

        return ax


    def _draw(self, obj, ax):
        if self._iterate:
            for item in obj:
                self._method(item, ax)
        elif self._iterate_df_columns:
            if not isinstance(obj,pd.DataFrame):
                raise ValueError('Value expect to be pd.Dataframe to do iterate over its columns')
            for c in obj.columns:
                item = obj[c]
                self._method(item, ax)
        else:
            self._method(obj, ax)


    def on(self, ax: Axes) -> 'FluentPlot':
        if ax is not None:
            self._default_ax = ax
        return self


    def size(self, width, height):
        self._default_fig_size = (width, height)
        return self


    def labels(self, title: Optional[str] = None, xlabel: Optional[str] = None,
               ylabel: Optional[str] = None) -> 'FluentPlot':
        self._title = title
        self._xlabel = xlabel
        self._ylabel = ylabel
        return self


    def with_legend(self) -> 'FluentPlot':
        self._with_legend = True
        return self


    def tune(self, method: Callable[[Axes], Any]) -> 'FluentPlot':
        self._post_processing.append(method)
        return self


    def call(self, method: Callable[[Any,Axes],Any]) -> 'FluentPlot':
        self._method = method
        return self

    def iterate(self, iterate=True) -> 'FluentPlot':
        self._iterate = iterate
        return self
    

    def iterate_df_columns(self, iterate_df_columns=True) -> 'FluentPlot':
        self._iterate_df_columns = iterate_df_columns
        return self




from yo_ds__tests.common import *
import seaborn
import math

class TestFluentPlot(TestCase):
    def test_1_simple(self):
        df = Query.en(range(10)).select(lambda z: Obj(x=z,y=z*z)).to_dataframe()
        return df.feed(FluentPlot().call(lambda item,ax: ax.plot(item.x,item.y)).labels('Simplest fluent plot'))

    def test_2_3rdparty(self):
        df = Query.combinatorics.grid(x=range(5),y=range(5)).select(lambda z: z.update(z=z.x+z.y)).to_dataframe().pivot_table(columns='x',index='y',values='z')
        return df.feed(FluentPlot().call(lambda item, ax: seaborn.heatmap(data=item,ax=ax)).labels('Fluent plot over 3rd party member from seaborn'))

    def test_3_tuning(self):
        df = Query.en(range(10)).select(lambda z: Obj(x=z, y=z * z)).to_dataframe()
        return df.feed(FluentPlot()
                       .call(lambda item,ax: ax.plot(item.x,item.y,label='label'))
                       .labels('Plot with cosmetic tuning','X axis','Y axis')
                       .with_legend()
                       .tune(lambda ax: ax.tick_params(axis='x', rotation=45))
                       )


    def test_4_groupby_plot(self):
        df = Query.combinatorics.grid(a=range(5),x=range(10)).select(lambda z: z.update(y=z.x*z.a)).to_dataframe()
        return df.groupby('a').feed(FluentPlot()
                                    .call(lambda gr, ax: ax.plot(gr[1].x, gr[1].y, label = gr[0]))
                                    .labels('Plot built for groupby')
                                    .iterate()
                                    .with_legend())


    def test_5_df_columns_plot(self):
        df = Query.combinatorics.grid(x=range(5), y=range(5)).select(lambda z: z.update(z=z.x + z.y)).to_dataframe().pivot_table(columns='x', index='y', values='z')
        return df.feed(FluentPlot()
                       .call(lambda item,ax: ax.plot(item.index, item, label = item.name))
                       .iterate_df_columns()
                       .labels('Plot build for dataframe columns')
                       .with_legend())


    def test_6_plots_on_different_axes(self):
        df = Query.en(range(10)).select(lambda z: Obj(x=z, pow=z*z, sqrt=math.sqrt(z))).to_dataframe().set_index('x')
        (Query
         .en(df.columns)
         .feed(fluq.with_plots(columns=2))
         .foreach(lambda p: df[p.item].feed(FluentPlot()
                                            .call(lambda item,ax: ax.plot(item.index, item))
                                            .on(p.ax)
                                            )))

    def test_7_plot_from_series(self):
        series = pd.Series([5,4,3,2,1])
        series.feed(FluentPlot().call(lambda item, ax: item.plot(ax=ax)).size(20,10))


    def test_8_plot_from_df(self):
        df = Query.en(range(10)).select(lambda z: Obj(x=z, pow=z * z, sqrt=math.sqrt(z))).to_dataframe().set_index('x')
        df.feed(FluentPlot().call(lambda item,ax: item.plot(ax=ax)).iterate_df_columns())


from yo_fluq_ds._fluq import *
import yo_fluq_ds._push_queries.aggregations as _agg

class to_pickle_file(_agg.ToPickleFile):
    pass

class to_text_file(_agg.ToTextFile):
    pass

class to_zip_text_file(_agg.ToZipTextFile):
    pass

class to_zipped_folder(_agg.ToZipFolder):
    pass



import unittest
from yo_ds import *

class KrakenMethodsTests(unittest.TestCase):
    def test_invoke(self):
        obj = dict(ctor=dict,a=1,b=2)
        result = kraken.invoke(obj)
        self.assertDictEqual(dict(a=1,b=2),obj['instance'])
        self.assertEqual('dict',obj['name'])

    def test_extract_id(self):
        df = Query.combinatorics.grid(a=[1,2],b=[3,4],c=[5,6]).to_dataframe()
        df, ids = kraken.extract_id(df,'t','a','b')
        self.assertListEqual(['1/3','1/3','1/4','1/4','2/3','2/3','2/4','2/4'],list(df.t))
        self.assertListEqual(list(ids.a.astype(str)+"/"+ids.b.astype(str)),list(ids.t))




import numpy as np
from yo_fluq_ds import *
from sklearn.metrics import roc_curve
from matplotlib import pyplot as plt
from matplotlib.axes import Axes



def _inner_roc_optimal_threshold(curve):
    dsts = curve[0] ** 2 + (1 - curve[1]) ** 2
    argmin = np.argmin(dsts)
    val = curve[2][argmin]
    return curve[0][argmin], curve[1][argmin], val

def roc_optimal_threshold(y_true, y_pred):
    curve = roc_curve(y_true, y_pred)
    _, __, val = _inner_roc_optimal_threshold(curve)
    return val



from yo_ds import *
from unittest import TestCase


def path(*args):
    folder = FileIO.find_root_folder('yo.root')
    return os.path.abspath(os.path.join(folder, 'tests', 'extensions', 'test_plots_extensions', *args))


from ._common import *
from sklearn.metrics import roc_curve
import numpy as np


def _inner_roc_optimal_threshold(curve):
    dsts = curve[0] ** 2 + (1 - curve[1]) ** 2
    argmin = np.argmin(dsts)
    val = curve[2][argmin]
    return curve[0][argmin], curve[1][argmin], val

def roc_optimal_threshold(y_true, y_pred):
    curve = roc_curve(y_true, y_pred)
    _, __, val = _inner_roc_optimal_threshold(curve)
    return val

def roc_plot(true,predicted, name=None, number_fmt='{:.2f}',ax=None):
    ax = default_ax(ax)
    curve = roc_curve(true,predicted)
    ax.plot(curve[0], curve[1],label=name)
    ax.set_xlabel('TPR')
    ax.set_ylabel('FPR')
    x, y, val = _inner_roc_optimal_threshold(curve)
    ax.scatter(x,y)
    ax.text(x,y,number_fmt.format(val),ha='left',va='top')



from yo_ds__tests.common import *

def _get_df():
    df = Query.combinatorics.grid(x=range(5), a=range(5)).select(lambda z: z.update(y=z.x + z.a, err=z.a/2)).to_dataframe()
    return df

class TestGrBar(TestCase):
    def test_1_simple(self):
        df = _get_df()
        grbar_plot(df,'y','a','x')

    def test_2_error(self): # TODO: does not work
        df = _get_df()
        grbar_plot(df,'y','a','x','err')


    def test_3_value_caption(self):
        df = _get_df()
        grbar_plot(df,'y','a','x','err',value_format='{0:.2f}±{1:.2f}')

    def test_4_custom_caption(self):
        df = _get_df()
        df = df.assign(caption=df.y)
        grbar_plot(df,'y','a','x',None,'caption')


    def test_5_witout_groups(self):
        df= _get_df()
        df = df.loc[df.a==1]
        grbar_plot(df,'y','x',None)

    def test_6_without_colors(self):
        df = _get_df()
        df = df.loc[df.a==1]
        grbar_plot(df,'y',None,'x')

    def test_7_orient_h(self):
        df = _get_df()
        grbar_plot(df,'y','a','x',orient='h')

    def test_8_confints(self):
        df = _get_df()
        df = df.assign(confint=df.apply(lambda z: pd.Interval(z.y-z.err,z.y+z.err),axis=1))
        grbar_plot(df,'confint','a','x')


from matplotlib import pyplot as plt
from matplotlib.axes import Axes
import pandas as pd
import numpy as np
from typing import *
from yo_fluq_ds._fluq._common import *


def default_ax(ax):
    if ax is None:
        _, ax = plt.subplots(1,1)
    return ax




class mb_call:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self._with_try_catch = False

    def try_catch(self):
        self._with_try_catch = True
        return self

    def _invoke(self, value):
        if not self._with_try_catch:
            return (value(*self.args, **self.kwargs), False)
        else:
            try:
                return (value(*self.args, **self.kwargs), False)
            except:
                return (None,True)



def maybe(value, *args, default = None):

    for action in args:

        if value is None:
            break

        elif callable(action):
            value = action(value)

        elif isinstance(action,str):
            try:
                value = getattr(value,action)
            except:
                value = None

        elif isinstance(action,mb_call):
            inv_result = action._invoke(value)
            value = inv_result[0]

        elif isinstance(action,list):

            if len(action)<1 or len(action)>3:
                raise ValueError("Only 1-, 2- and 3-dimentional indexing is supported") #pragma: no cover
            try:
                if len(action)==1:
                    value = value[action[0]]
                elif len(action)==2:
                    value = value[action[0],action[1]] #No star-expression for indexing,
                elif len(action)==3:
                    value = value[action[0], action[1],action[2]]
                else:
                    pass
            except:
                value = None

    if value is None:
        return default

    return value




from yo_ds import *
from unittest import TestCase
from IPython.display import HTML

class TestYoMisc(TestCase):
    def test_notebook_true(self):
        self.assertIsInstance(
            notebook_printable_version(True),
            HTML
        )

    def test_notebook_false(self):
        self.assertIsNone(
            notebook_printable_version(False)
        )

    def test_diff_set(self):
        s = diffset({1,2,3},{2,3,4})
        self.assertListEqual([3,3,1,1,2,False,False,False],list(s.values))



from setuptools import setup, find_packages

def readme():
      with open('README.md') as file:
            return file.read()

setup(name='yo_ds',
      version='1.1.7',
      description='Personal library with data science tools',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers = [
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      url='http://github.com/okulovsky/yo_ds',
      author='Yuri Okulovsky',
      author_email='yuri.okulovsky@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'pandas',
            'matplotlib',
            'numpy',
            'tqdm',
            'seaborn',
            'pyaml',
            'jsonpickle',
            'ipython',
            'ipywidgets',
            'yo_fluq_ds==1.1.7'
      ],
      include_package_data = True,
      zip_safe=False)


from yo_fluq_ds import *
import os
from collections import OrderedDict
import copy
import sys
from typing import *
import pandas as pd
import numpy as np

EXTENTION = '.kraken.pkl'


def unwrap(obj, prefix = '', result=None):
    if result is None:
        result = OrderedDict()
    if isinstance(obj, dict):
        for key, value in obj.items():
            unwrap(value,
                   prefix+('_' if prefix!='' else '')+str(key),
                   result)
    else:
        result[prefix] = obj
    return result


class IterationStatus(OrderedEnum):
    Failed = 0
    Success = 1
    Skipped = 2



class IterationResult:
    def __init__(self,
                 status : IterationStatus,
                 result: Optional[Any] = None,
                 condition: Optional[Any] = None,
                 exception_info: Optional[Any] = None
                 ):
        self.status = status
        self.result = result
        self.condition = condition
        self.exception_info = exception_info


class _IterClass:
    def __init__(self, method, cache_to_folder, handle_exception_callback):
        self.method = method
        self.cache_to_folder = cache_to_folder
        self.handle_exception_callback = handle_exception_callback


    def process(self, config: Any) -> Optional[IterationResult]:
        if self.cache_to_folder is not None:
            cache_path = os.path.join(self.cache_to_folder, str(config['iteration']) + EXTENTION)
            if os.path.isfile(cache_path):
                return IterationResult(IterationStatus.Skipped)

        condition = copy.deepcopy(config)

        value = None
        status = IterationStatus.Success
        exception = None

        if self.handle_exception_callback is None:
            value = self.method(**condition)
        else:
            try:
                value = self.method(**condition)
            except:
                exception = sys.exc_info()
                self.handle_exception_callback(condition)
                status = IterationStatus.Failed

        result = IterationResult(status,value,condition,exception)

        if self.cache_to_folder is not None:
            cache_path = os.path.join(self.cache_to_folder, str(config['iteration']) + EXTENTION)
            with open(cache_path, 'wb') as file:
                pickle.dump(result, file)
            return IterationResult(IterationStatus.Success)
        else:
            return result


def release(method: Callable,
            configs: Iterable,
            pandas_extractor : Callable[[Any,Any],pd.DataFrame] = lambda df, _ : df,
            shuffle: Union[bool, int] = False,
            handle_exception_callback: Optional[Callable[[Any],None]] = None,
            special_iterations: Optional[Iterable[int]] = None,
            cache_to_folder: Optional[str] = None,
            with_tqdm: bool = True,
            parallel_kwargs: Optional[Dict] = None) -> Union[pd.DataFrame,List[IterationResult]]:
    configs = (Query
               .en(configs)
               .select(copy.deepcopy)
               .with_indices()
               .foreach_and_continue(lambda z: z.item.update(iteration=z.index))
               .select(lambda z: z.item)
               .to_list()
               )

    if special_iterations is not None:
        special_iterations = list(special_iterations)
        configs = np.array(configs)
        configs = list(configs[special_iterations])

    configs = Query.en(configs).feed(fluq.shuffle(shuffle)).to_list()


    if cache_to_folder:
        os.makedirs(cache_to_folder, exist_ok=True)
        existed = Query.folder(cache_to_folder).select(lambda z: z.name.replace(EXTENTION, '')).select(int).to_set()
        configs = Query.en(configs).where(lambda z: z['iteration'] not in existed).to_list()
        if len(configs)==0:
            return []


    executor = _IterClass(method,cache_to_folder, handle_exception_callback)

    output = [] # type: Optional[List[IterationResult]]

    if parallel_kwargs is None:
        output = (
            Query
            .en(configs)
            .select(executor.process)
        )
    else:
        output = (
            Query
            .en(configs)
            .parallel_select(executor.process,**parallel_kwargs)
        )

    if with_tqdm:
        output = output.feed(fluq.with_progress_bar(total=len(configs)))

    output = output.to_list()

    if cache_to_folder:
        return output
    else:
        if pandas_extractor is None:
            return output
        else:
            return combine_result(output,pandas_extractor)


def load_results_iter(folder: str, files=None) -> Iterable[IterationResult]:
    if files is None:
        files = [file for file in os.listdir(folder) if file.endswith(EXTENTION)]
    else:
        files = [str(f)+EXTENTION for f in files]
    for file in Query.en(files).feed(fluq.with_progress_bar()):
        with open(os.path.join(folder, file), 'rb') as f:
            yield pickle.load(f)


def load(folder,  pd_extractor = lambda df, _ : df, exclude_config_fields = None) -> Union[List[IterationResult],pd.DataFrame]:
    loading = load_results_iter(folder)
    if pd_extractor is None:
        return list(loading)
    else:
        return combine_result(loading,
            pd_extractor,
            exclude_config_fields
        )

def combine_result(results, pd_extractor = lambda df, _ : df, exclude_config_fields = None) -> pd.DataFrame:
    output = None
    conditions = []
    for result in results:
        if result.status!=IterationStatus.Success:
            continue
        condition = unwrap(result.condition)
        pd_result = pd_extractor(result.result, condition)
        pd_result['iteration'] = condition['iteration']
        if output is None:
            output = pd_result
        else:
            output = output.append(pd_result)
        if exclude_config_fields is not None:
            for field in exclude_config_fields:
                del condition[field]
        conditions.append(condition)

    conditions = pd.DataFrame(conditions)
    output = output.merge(conditions.set_index('iteration'), left_on='iteration',right_index=True)
    output = output.reset_index(drop=True)
    return output


def invoke(factory):
    kwargs = copy.deepcopy(factory)
    del kwargs['ctor']
    factory['instance'] = factory['ctor'](**kwargs)
    factory['name'] = factory['ctor'].__name__
    return factory



def extract_id(df, name, *args):
    id_column = None
    args = list(args)
    for column in args:
        addition = np.where(df[column].isnull(), '', df[column].astype(str))
        if id_column is None:
            id_column = addition
        else:
            id_column = id_column + np.where((addition != '') & (id_column != ''), '/', '') + addition

    id_column_name = name
    df = df.assign(**{id_column_name: id_column})
    ids = df[[id_column_name] + args].drop_duplicates()
    df = df.drop(args, axis=1)
    return df, ids


from ._common import *


def _assign(df: pd.DataFrame, items, name: str, smth: Union[str,pd.Series]):
    if smth is None:
        raise ValueError('Parameter for {0} cannot be None'.format(name))
    elif isinstance(smth, pd.Series):
        if len(smth) != df.shape[0]:
            raise ValueError('Length of a series, given as {0}, does not match dataframe'.format(name))
        column = smth
    else:
        column = df[smth]

    for item, value in zip(items,column):
        setattr(item,name,value)

def _get_name(smth: Union[None,str,pd.Series]):
    if smth is None:
        return None
    elif isinstance(smth,pd.Series):
        return smth.name
    else:
        return str(smth)



class _DbarItem:
    def __init__(self):
        self.value = 0
        self.color = ''
        self.group = ''
        self.error = None
        self.caption = None

        self.low_value = None
        self.high_value = 0

        self.color_index = 0
        self.group_index = 0

        self.x = 0

    def finalize(self, value_format, color_indices, group_indices):
        if isinstance(self.value, pd.Interval):
            self.error = self.value.mid - self.value.left
            self.value = self.value.mid

        if value_format is not None and self.caption is None:
            if self.error is not None:
                self.caption = value_format.format(self.value,self.error)
            else:
                self.caption = value_format.format(self.value)

        if self.error is None:
            self.low_value = None
            self.high_value = self.value
        else:
            self.low_value = self.value - self.error
            self.high_value = self.value + self.error

        self.color_index = color_indices[self.color]
        self.group_index = group_indices[self.group]

        self.group_width = 1
        self.group_draw_width = 0.8
        self.color_width = self.group_draw_width / (len(color_indices))

        self.x = self.group_index * self.group_width + self.color_index * self.color_width




def _plot_dbar_vertical(ax, items: List[_DbarItem], group_label, value_label):
    gitems = Query.en(items).group_by(lambda z: z.color_index).select(list).to_list()
    for color_items in gitems:
        rects = ax.bar(
            Query.en(color_items).select(lambda z: z.x).to_list(),
            Query.en(color_items).select(lambda z: z.high_value).to_list(),
            width=color_items[0].color_width,
            label=color_items[0].color
        )

        with_low = Query.en(color_items).where(lambda z: z.low_value is not None).to_list()
        if len(with_low)>0:
            ax.bar(
                Query.en(with_low).select(lambda z: z.x).to_list(),
                Query.en(with_low).select(lambda z: z.low_value).to_list(),
                width=with_low[0].color_width,
                color='white',
                alpha=0.6
            )

        for item, rect in zip(color_items, rects):
            if item.caption is None:
                continue
            ax.text(
                rect.get_x() + rect.get_width() / 2.,
                rect.get_height(),
                item.caption,
                ha='center', va='bottom')

    labels = Query.en(items).distinct(lambda z: z.group_index).to_series(lambda z: z.group,lambda z: z.group_index)
    ax.xaxis.set_ticks([x * items[0].group_width + items[0].group_draw_width / 2 - items[0].color_width / 2 for x in labels.index])
    ax.xaxis.set_ticklabels(labels)
    if group_label is not None:
        ax.set_xlabel(group_label)
    ax.set_ylabel(value_label)
    return ax


def _plot_dbar_horizontal(ax, items: List[_DbarItem], group_label, value_label):
    gitems = Query.en(items).group_by(lambda z: z.color_index).select(list).to_list()
    for color_items in gitems:
        rects = ax.barh(
            Query.en(color_items).select(lambda z: z.x).to_list(),
            Query.en(color_items).select(lambda z: z.high_value).to_list(),
            height=color_items[0].color_width,
            label=color_items[0].color
        )
        with_low = Query.en(color_items).where(lambda z: z.low_value is not None).to_list()
        if len(with_low) > 0:
            ax.barh(
                Query.en(with_low).select(lambda z: z.x).to_list(),
                Query.en(with_low).select(lambda z: z.low_value).to_list(),
                height=with_low[0].color_width,
                color='white',
                alpha=0.6
            )

        for item, rect in zip(color_items, rects):
            if item.caption is None:
                continue
            ax.text(
                rect.get_width(),
                rect.get_y() + rect.get_height() / 2.,
                item.caption,
                ha='left', va='center')

    labels = Query.en(items).distinct(lambda z: z.group_index).to_series(lambda z: z.group, lambda z: z.group_index)
    ax.yaxis.set_ticks(
        [x * items[0].group_width + items[0].group_draw_width / 2 - items[0].color_width / 2 for x in labels.index])
    ax.yaxis.set_ticklabels(labels)
    if group_label is not None:
        ax.set_ylabel(group_label)
    ax.set_xlabel(value_label)
    return ax




def grbar_plot(
         df: pd.DataFrame,
         value_column: Union[str, pd.Series],
         color_column: Union[str, pd.Series, None] = None,
         group_column: Union[str, pd.Series, None] = None,
         error_column: Union[str, pd.Series, None] = None,
         caption_column: Union[str, pd.Series, None] = None,
         value_format: Optional[str] = None,
         ax = None,
         orient = 'v'
         ):
    items = [_DbarItem() for _ in range(df.shape[0])]

    _assign(df, items, 'value', value_column)
    if color_column is not None:
        _assign(df, items, 'color', color_column)
    else:
        if group_column is None:
            raise ValueError('At least one of `color_column` or `group_column` must be not None')
    if group_column is not None:
        _assign(df, items, 'group', group_column)
    if error_column is not None:
        _assign(df, items, 'error', error_column)
    if caption_column is not None:
        _assign(df, items, 'caption', caption_column)

    color_indices = (Query
                     .en(items)
                     .select(lambda z: z.color)
                     .distinct()
                     .with_indices()
                     .to_dictionary(lambda z: z.value, lambda z: z.key)
                     )

    group_indices = (Query
                     .en(items)
                     .select(lambda z: z.group)
                     .distinct()
                     .with_indices()
                     .to_dictionary(lambda z: z.value, lambda z: z.key)
                     )

    Query.en(items).foreach(lambda z: z.finalize(value_format, color_indices, group_indices))

    group_label = _get_name(group_column)
    value_label = _get_name(value_column)

    if orient=='v':
        ax = _plot_dbar_vertical(default_ax(ax), items, group_label, value_label)
    else:
        ax = _plot_dbar_horizontal(default_ax(ax), items, group_label, value_label)

    return ax








import os
import json
import pickle
from yo_fluq_ds._misc.io import FileIO as IO
from yo_ds import *



def find_root_folder(root_file_name):
    root_path = './'
    for i in range(10):
        if os.path.isfile(root_path+root_file_name):
            return root_path
        root_path+='../'
    raise ValueError("Cound't find the root {1}. The current directory is {0}".format(os.path.abspath('.'), root_file_name))

def load_json(filename, as_obj=False):
    result = None
    with open(filename) as file:
        result = json.load(file)
    if as_obj:
        result = Obj.create(result)
    return result

def save_json(filename, obj):
    with open(filename,'w') as file:
        json.dump(obj,file,indent=1)

def load_pkl(filename):
    with open(filename,'rb') as file:
        return pickle.load(file)


def save_pkl(filename, obj):
    with open(filename,'wb') as file:
        pickle.dump(obj,file)





from IPython.display import HTML

def notebook_printable_version(finalize):
    if finalize:
        return HTML('''<script>
        code_show=true; 
        function code_toggle() {
         if (code_show){
         $('div.input').hide();
         $("div[class='prompt output_prompt']").css('visibility','hidden');

         } else {
         $('div.input').show();
         $("div[class='prompt output_prompt']").css('visibility','visible');
         }
         code_show = !code_show
        } 
        $( document ).ready(code_toggle);
        </script>
        <a href="javascript:code_toggle()">Automatically generated report</a>.''')
    else:
        return None

import pandas as pd

def diffset(set1, set2, name1='First',name2='Second'):
    set1=set(set1)
    set2=set(set2)
    return pd.Series([
    len(set1),
    len(set2),
    len(set1-set2),
    len(set2-set1),
    len(set1.intersection(set2)),
    set1==set2,
    set1.issubset(set2),
    set2.issubset(set1),
    ],index=[name1,name2,'{0}-{1}'.format(name1,name2),'{0}-{1}'.format(name2,name1),"Intersection","Match",'LeftIsSubset','RightIsSubset']
)
